/*
 * $URL$
 * 
 * Copyright (C) 2022 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * $Author$
 *   $Date$
 *    $Rev$
 *
 * (description of this file here)
 *
 */
#include "image_struct.h"
#include "constants.h"
#include "read_edr.h"
#include <stdlib.h>
#include "decompand.h"
#include "bias_subtract.h"
#include "gain_correction.h"
#include "dark_subtract.h"
#include "flat_field.h"
#include "radiance.h"
#include "write_cube.h"
#include <stdbool.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include "config_path.h"
#include <errno.h>
#include "special_pixels.h"

// XXX maybe do something better here
#define VERSION "$URL$\n$Id$"

void usage(int argc, char*argv[]);
void usage(int argc, char*argv[]) {
  printf("Usage: %s [OPTIONS...] -i EDR_FILE -o CDR_FILE\n", argv[0]);
  printf("\n");
  printf("This tool takes a ShadowCam EDR and performs all calibration steps in order.\n");
  printf("The output is a 32-bit ISIS cube.\n");
  printf("\n");
  printf("Options:\n");
  printf("  -i FILE, --infile=FILE      The EDR cube to input.\n");
  printf("  -o FILE, --outfile=FILE     The CDR cube to output.\n");
  printf("  -w, --write-intermediate    Write out individual cubes after each step.\n");
  printf("                              (default is final only.)\n");
  printf("  -k, --keep-bias             Keep bias pixels in the output image.\n");
  printf("                              (default is remove bias pixels.)\n");
  printf("  -S, --nospecial             Don't replace 0/255 with LIS/HIS.\n");
  printf("                              (default is to replace special pixels.\n");
  printf("  -m, --median                Use the median bias value for bias subtraction.\n");
  printf("                              (default is to use the mean.)\n");
  printf("  -s N, --step=N              Which step to stop calibrating after.\n");
  printf("                              1: Decompanding\n");
  printf("                              2: Bias Subtraction\n");
  printf("                              3: Gain Normalization\n");
  printf("                              4: Dark Subtraction\n");
  printf("                              5: Flat Field\n");
  printf("                              6: Radiance (default)\n");
  printf("  -h, --help                  Show this help.\n");
  printf("  -V, --version               Print program version.\n");
}

int main(int argc, char *argv[]) {
  int opt=0;

  char *infile = NULL;
  char *outfile = NULL;
  bool write_intermediate = false;
  bool keep_bias = false;
  bool nospecial = false;
  uint8_t step = 255;
  bool use_median = false;

  static struct option long_options[] = {
    {"infile", required_argument, 0, 'i'},
    {"outfile", required_argument, 0, 'o'},
    {"write-intermediate", no_argument, 0, 'w'},
    {"keep-bias", no_argument, 0, 'k'},
    {"step", required_argument, 0, 's'},
    {"nospecial", no_argument, 0, 'S'},
    {"median", no_argument, 0, 'm'},
    {"help", no_argument, 0, 'h'},
    {"version", no_argument, 0, 'V'},
    {0,0,0,0}};

  int long_index = 0;

  while((opt = getopt_long(argc, argv, "i:o:wks:SmhV", long_options, &long_index)) != -1) {
    char *tail;

    switch(opt) {
      case 'i':
        infile = strdup(optarg);
        break;
      case 'o':
        outfile = strdup(optarg);
        break;
      case 'w':
        write_intermediate = true;
        break;
      case 'k':
        keep_bias = true;
        break;
      case 's':
        errno = 0;
        step = (uint8_t)strtoul(optarg, &tail, 10);
        if (errno || *tail) {
          printf("ERROR: step '%s' does not appear to be a valid integer.\n\n", optarg);
          usage(argc, argv);
          exit(1);
        }
        break;
      case 'S':
        nospecial = true;
        break;
      case 'm':
        use_median = true;
        break;
      case 'h':
        usage(argc, argv);
        exit(0);
      case 'V':
        printf("%s\n", VERSION);
        exit(0);
      default:
        usage(argc, argv);
        exit(1);
    }
  }

  if (infile == NULL || outfile == NULL) {
    printf("ERROR: Both infile and outfile must be specified.\n\n");
    usage(argc, argv);
    exit(1);
  }

  config_path();

  struct shadowcam_image *image = init_shadowcam_image(infile);

  read_edr(image, nospecial);
  if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "companded");

  if (step >= 1) {
    decompand(image);
    if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "decompanded");
  }

  if (step >= 2) {
    bias_subtract(image, use_median);
    if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "bias_subtract");
  }

  if (step >= 3) {
    gain_correction(image);
    if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "gain_correct");
  }

  if (step >= 4) {
    dark_subtract(image);
    if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "dark_subtract");
  }

  if (step >= 5) {
    flat_field(image);
    if (write_intermediate) write_cube_suffix(keep_bias, outfile, image, "flat_field");
  }

  if (step >= 6)
    radiance(image);

  write_cube(keep_bias, outfile, image);

  free_shadowcam_image(image);

  return 0;
}
